# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""

Best-effort crypto primitives for PyBrowserID.

If you have M2Crypto installed, this package will provide a nice fast
implementation of the RSKey and DSKey classes needed to do the crypto
work behind BrowserID.  If you don't, you'll get a very slow but very
portable pure-python version.

"""

try:
    from browserid.crypto.m2 import Key, RSKey, DSKey
except ImportError:
    from browserid.crypto.fallback import Key, RSKey, DSKey  # NOQA



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""

Error classes for PyBrowserID.

"""


class Error(Exception):
    """Base error class for all PyBrowserID exceptions."""
    pass


class ConnectionError(Error):
    """Error raised when PyBrowserID fails to connect to a remote server."""
    description = 'Failed to connect to remote server.'


class TrustError(Error):
    """Base error class when assertions are well-formed but untrusted."""
    description = 'Untrusted assertion.'


class InvalidSignatureError(TrustError):
    """Error raised when PyBrowserID encounters an invalid signature."""
    description = 'Signature is invalid.'


class InvalidIssuerError(TrustError):
    """Error raised when a cert is from an invalid/untrusted issuer."""
    description = 'Certificate issued by invalid or untrusted issuer.'


class ExpiredSignatureError(TrustError):
    """Error raised when PyBrowserID encounters an expired signature."""
    description = 'Signature is expired.'


class AudienceMismatchError(TrustError):
    """Error raised when the audience does not match."""
    description = 'Audience does not match server.'


class UnsupportedCertChainError(TrustError):
    """The spec for multi-cert chains is in flux; we don't support them yet."""
    description = 'Multiple-certificate chains not supported.'



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import re
import sys
import time

from browserid import jwt
from browserid.verifiers import Verifier
from browserid.supportdoc import SupportDocumentManager
from browserid.utils import unbundle_certs_and_assertion
from browserid.errors import (InvalidSignatureError,
                              ExpiredSignatureError,
                              UnsupportedCertChainError)


if sys.version_info > (3,):
    basestring = (str,)

VALID_EMAIL = re.compile(r"^([\w!#$%&'*+/=?^`{|}~.\-]+)@" + \
                         r"([\w.\-]+(:[0-9]{1,5})?)$")

RESERVED_CLAIM_NAMES = set((
  "iss",
  "sub",
  "aud",
  "exp",
  "nbf",
  "iat",
  "jti",
  "public-key",
  "pubkey",
  "principal"
))


class LocalVerifier(Verifier):
    """Class for local verification of BrowserID identity assertions.

    This class implements the logic for verifying identity assertions under
    the Verified Email Protocol.  Pass a BrowserID assertion token to the
    verify() method and let it work its magic.
    """

    def __init__(self, audiences=None, trusted_secondaries=None,
                 supportdocs=None, warning=False):
        super(LocalVerifier, self).__init__(audiences)
        self.trusted_secondaries = trusted_secondaries
        self.supportdocs = supportdocs or SupportDocumentManager()

        if warning:
            # We used to emit a warning about the assertion format
            # being subject to change.  That's no longer true, but
            # we still accept the `warning` arg for compatibility.
            pass

    def parse_jwt(self, data):
        return jwt.parse(data)

    def verify(self, assertion, audience=None, now=None):
        """Verify the given BrowserID assertion.

        This method parses a BrowserID identity assertion, verifies the
        bundled chain of certificates and signatures, and returns the
        extracted email address and audience.

        If the 'audience' argument is given, it first verifies that the
        audience of the assertion matches the one given.  This can help
        avoid doing lots of crypto for assertions that can't be valid.
        If you don't specify an audience, you *MUST* validate the audience
        value returned by this method.

        If the 'now' argument is given, it is used as the current time in
        milliseconds.  This lets you verify expired assertions, e.g. for
        testing purposes.
        """
        if now is None:
            now = int(time.time() * 1000)

        # This catches KeyError and turns it into ValueError.
        # It saves having to test for the existence of individual
        # items in the various assertion payloads.
        try:
            # Check the audience against the given value, or the wildcards.
            self.check_audience(assertion, audience)

            # Grab the assertion, check that it has not expired.
            # No point doing all that crypto if we're going to fail out anyway.
            certificates, assertion = unbundle_certs_and_assertion(assertion)
            if len(certificates) > 1:
                raise UnsupportedCertChainError("too many certs")
            assertion = self.parse_jwt(assertion)
            if assertion.payload["exp"] < now:
                raise ExpiredSignatureError(assertion.payload["exp"])

            # Parse out the list of certificates.
            certificates = [self.parse_jwt(c) for c in certificates]

            # Extract the email, and the hostname of its provider.
            email = certificates[-1].payload["principal"]["email"]
            match = VALID_EMAIL.match(email)
            if match is None:
                raise ValueError("invalid email in assertion")
            provider = match.group(2)

            # Check that the root issuer is trusted.
            # No point doing all that crypto if we're going to fail out anyway.
            root_issuer = certificates[0].payload["iss"]
            if not self.is_trusted_issuer(provider, root_issuer):
                msg = "untrusted root issuer: %s" % (root_issuer,)
                raise InvalidSignatureError(msg)

            # Verify the entire chain of certificates.
            cert = self.verify_certificate_chain(certificates, now=now)

            # Check the signature on the assertion.
            if not self.check_token_signature(assertion, cert):
                raise InvalidSignatureError("invalid signature on assertion")
        except KeyError:
            raise ValueError("Malformed JWT")
        # Looks good!
        res = {
          "status": "okay",
          "audience": assertion.payload["aud"],
          "email": email,
          "issuer": root_issuer,
        }
        idpClaims = extract_extra_claims(certificates[-1])
        if idpClaims:
            res["idpClaims"] = idpClaims
        userClaims = extract_extra_claims(assertion)
        if userClaims:
            res["userClaims"] = userClaims
        return res

    def is_trusted_issuer(self, hostname, issuer):
        """Check whether the issuer is trusted for a given hostname."""
        return self.supportdocs.is_trusted_issuer(hostname, issuer,
                                                  self.trusted_secondaries)

    def check_token_signature(self, data, cert):
        """Check for a valid signature on the given JWT."""
        return data.check_signature(cert.payload["public-key"])

    def verify_certificate_chain(self, certificates, now=None):
        """Verify a signed chain of certificates.

        This function checks the signatures on the given chain of JWT
        certificates.  It looks up the public key for the issuer of the
        first certificate, then uses each certificate in turn to check the
        signature on its successor.

        If the entire chain is valid then to final certificate is returned.
        """
        if not certificates:
            raise ValueError("chain must have at least one certificate")
        if now is None:
            now = int(time.time() * 1000)
        root_issuer = certificates[0].payload["iss"]
        root_key = self.supportdocs.get_key(root_issuer)
        current_key = root_key
        for cert in certificates:
            if cert.payload["exp"] < now:
                raise ExpiredSignatureError("expired certificate in chain")
            if not cert.check_signature(current_key):
                raise InvalidSignatureError("bad signature in chain")
            current_key = cert.payload["public-key"]
        return cert


def extract_extra_claims(jwt):
    """Return a dict of any non-standard claims in the given JWT."""
    claims = {}
    for claim in jwt.payload:
        if claim not in RESERVED_CLAIM_NAMES:
            claims[claim] = jwt.payload[claim]
    principal = jwt.payload.get("principal", None)
    if principal is not None and not isinstance(principal, basestring):
        for claim in principal:
            if claim not in RESERVED_CLAIM_NAMES and claim != "email":
                claims.setdefault(claim, principal[claim])
    return claims



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""

Network-related utility functions for PyBrowserID.

"""

import socket

import requests
from requests.exceptions import RequestException

from browserid.errors import ConnectionError


def get(url, verify=True):
    """Fetch the specified URL with a GET request."""
    return request("GET", url, verify=verify)


def post(url, data={}, verify=True):
    """Fetch the specified URL with a POST request."""
    return request("POST", url, data=data, verify=verify)


def request(method, url, **kwds):
    """Make an HTTP request to the given URL."""
    try:
        return requests.request(method, url, **kwds)
    except (RequestException, socket.error) as e:
        msg = "Failed to %s %s. Reason: %s" % (method, url, str(e))
        raise ConnectionError(msg)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

from browserid.utils import encode_bytes, decode_bytes
from browserid.utils import encode_json_bytes, decode_json_bytes
from browserid.utils import get_assertion_info, u
from browserid.tests.support import unittest


class TestUtils(unittest.TestCase):

    def test_encode_decode_bytes(self):
        self.assertEquals(b"HELLO", decode_bytes(encode_bytes(b"HELLO")))
        self.assertEquals(b"HELLO", decode_bytes(encode_bytes(u("HELLO"))))
        self.assertRaises(ValueError, decode_bytes, u("\N{SNOWMAN}"))
        self.assertRaises(ValueError, decode_bytes, "A===")

    def test_encode_decode_json_bytes(self):
        obj = {"hello": "world"}
        self.assertEquals(obj, decode_json_bytes(encode_json_bytes(obj)))
        self.assertRaises(ValueError,
                          decode_json_bytes, encode_bytes("NOJSON4U"))
        self.assertRaises(ValueError,
                          decode_json_bytes, encode_bytes("42"))
        self.assertRaises(ValueError,
                          decode_json_bytes, encode_bytes("[1, 2, 3]"))
        self.assertRaises(ValueError, encode_json_bytes, 42)
        self.assertRaises(ValueError, encode_json_bytes, [1, 3, 3])

    def test_get_assertion_info(self):
        assertion = """
        eyJjZXJ0aWZpY2F0ZXMiOlsiZXlKaGJHY2lPaUpTVXpFeU9DSjkuZXlKcGMzTWlPaUppY
        205M2MyVnlhV1F1YjNKbklpd2laWGh3SWpveE16SXhPVFF4T1Rnek1EVXdMQ0p3ZFdKc2
        FXTXRhMlY1SWpwN0ltRnNaMjl5YVhSb2JTSTZJbEpUSWl3aWJpSTZJamd4TmpreE5UQTB
        OVGswTkRVek5EVTFPREF4TlRreU5Ea3hNemsyTkRFNE56RTJNVFUwTkRNNE5EWXdPREl6
        TXpBMU1USXlPRGN3TURRNE56TTFNREk1TURrek16a3lNRFkzTURFMU1qQTBORGd6TWpVM
        U56WXdOREE1TnpFeU9EYzNNVGswT1RVek1UQXdNVFEyTkRVek56TTJOakU0TlRVek5EY3
        hNakkxT0RreU16TTFPRFV4TWpZNU1EQXdOREF5TVRrMk9ERTBNRGtpTENKbElqb2lOalU
        xTXpjaWZTd2ljSEpwYm1OcGNHRnNJanA3SW1WdFlXbHNJam9pY25saGJrQnlabXN1YVdR
        dVlYVWlmWDAua19oaEtYMFRCVnUyX2szbV9uRDVOVWJfTktwX19PLTY1MW1CRUl3S1NZZ
        GlOenQwQm9WRkNEVEVueEhQTWJCVjJaejk0WDgtLVRjVXJidEV0MWV1S1dWdjMtNTFUOU
        xBZnV6SEhfekNCUXJVbmxkMVpXSmpBM185ZEhQeTMwZzRMSU9YZTJWWmd0T1Nva3MyZFE
        4ZDNvazlSUTJQME5ERzB1MDBnN3lGejE4Il0sImFzc2VydGlvbiI6ImV5SmhiR2NpT2lK
        U1V6WTBJbjAuZXlKbGVIQWlPakV6TWpFNU1qazBOelU0TWprc0ltRjFaQ0k2SW1oMGRIQ
        TZMeTl0ZVdaaGRtOXlhWFJsWW1WbGNpNXZjbWNpZlEuQWhnS2Q0eXM0S3FnSGJYcUNSS3
        hHdlluVmFJOUwtb2hYSHk0SVBVWDltXzI0TWdfYlU2aGRIMTNTNnFnQy1vSHBpS3BfTGl
        6cDRGRjlUclBjNjBTRXcifQ
        """.replace(" ", "").replace("\n", "").strip()
        data = get_assertion_info(assertion)
        self.assertEquals(data["principal"]["email"], "ryan@rfk.id.au")
        self.assertEquals(data["audience"], "http://myfavoritebeer.org")
        self.assertRaises(ValueError, get_assertion_info, "JUNK")
        self.assertRaises(ValueError, get_assertion_info, "X")
        self.assertRaises(ValueError, get_assertion_info, "\x00\x01\x02")
        bad_assertion = encode_json_bytes({"fake": "assertion"})
        self.assertRaises(ValueError, get_assertion_info, bad_assertion)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""

Crypto primitives implemented in pure python.

This file provides the "slow path" crypto implementation for PyBrowserID.
It implements everything in pure python, so it's very slow but very portable.

There is also a faster version built on M2Crypto, which should be picked up
automatically if you have that package installed.

"""

import os
from binascii import hexlify, unhexlify


class Key(object):
    """Generic base class for Key objects."""

    @classmethod
    def from_pem_data(cls, data=None, filename=None):
        """Alternative constructor for loading from PEM format data."""
        msg = "PEM data loading is not implemented for pure-python crypto."
        msg += "  Please install M2Crypto to access this functionality."
        raise NotImplementedError(msg)

    def to_pem_data(self):
        """Save the public key data to a PEM format string."""
        msg = "PEM data saving is not implemented for pure-python crypto."
        msg += "  Please install M2Crypto to access this functionality."
        raise NotImplementedError(msg)

    def verify(self, signed_data, signature):
        """Verify the given signature."""
        raise NotImplementedError

    def sign(self, data):
        """Sign the given data."""
        raise NotImplementedError


#  These constants are needed for encoding the name of the hash
#  algorithm into the RSA signature, per PKCS #1.
RSA_DIGESTINFO_HEADER = {
    "sha1": b"3021300906052b0e03021a05000414",
    "sha256": b"3031300d060960864801650304020105000420",
}


class RSKey(Key):
    """Generic base class for RSA key objects.

    Concrete subclasses should provide the DIGESTSIZE, HASHNAME and HASHMOD
    attributes.
    """

    # The size of the internal hex digest, in bytes.
    # This must equal the bit-length of the modulus "n" divided by 4.
    # The digest gets padded to this size to ensure that, when converted to
    # an integer, it will be of a similar magnitude to the modulus.
    DIGESTSIZE = None
    # The name and hashlib module to use for calculating the digest.
    HASHNAME = None
    HASHMOD = None

    def __init__(self, data):
        _check_keys(data, ("e", "n"))
        self.e = int(data["e"])
        self.n = int(data["n"])
        try:
            self.d = int(data["d"])
        except KeyError:
            self.d = None

    def verify(self, signed_data, signature):
        n, e = self.n, self.e
        m = int(hexlify(signature), 16)
        c = pow(m, e, n)
        digest = hex(c)[2:].rstrip("L").encode("ascii")
        padded_digest = digest.rjust(self.DIGESTSIZE, b"0")
        return padded_digest == self._get_digest(signed_data)

    def sign(self, data):
        n, e, d = self.n, self.e, self.d
        if not d:
            raise ValueError("private key not present")
        c = int(self._get_digest(data), 16)
        m = pow(c, d, n)
        return int2bytes(m)

    def _get_digest(self, data):
        digest = self.HASHMOD(data).hexdigest().encode("ascii")
        padded_digest = b"00" + RSA_DIGESTINFO_HEADER[self.HASHNAME] + digest
        padding_len = (self.DIGESTSIZE) - 4 - len(padded_digest)
        padded_digest = b"0001" + (b"f" * padding_len) + padded_digest
        return padded_digest


class DSKey(Key):
    """Generic base class for DSA key objects.

    Concrete subclasses should provide the BITLENGTH and HASHMOD attributes.
    """

    # The length of the signature to be produced, in bits.
    BITLENGTH = None
    # The hashlib module used to calculate the digest.
    HASHMOD = None

    def __init__(self, data):
        _check_keys(data, ("p", "q", "g", "y"))
        self.p = int(data["p"], 16)
        self.q = int(data["q"], 16)
        self.g = int(data["g"], 16)
        self.y = int(data["y"], 16)
        if "x" in data:
            self.x = int(data["x"], 16)
        else:
            self.x = None

    def verify(self, signed_data, signature):
        p, q, g, y = self.p, self.q, self.g, self.y
        signature = hexlify(signature)
        hexlength = self.BITLENGTH // 4
        signature = signature.rjust(hexlength * 2, b"0")
        if len(signature) != hexlength * 2:
            return False
        r = int(signature[:hexlength], 16)
        s = int(signature[hexlength:], 16)
        if r <= 0 or r >= q:
            return False
        if s <= 0 or s >= q:
            return False
        w = modinv(s, q)
        u1 = (int(self.HASHMOD(signed_data).hexdigest(), 16) * w) % q
        u2 = (r * w) % q
        v = ((pow(g, u1, p) * pow(y, u2, p)) % p) % q
        return (v == r)

    def sign(self, data):
        p, q, g, y, x = self.p, self.q, self.g, self.y, self.x
        if not x:
            raise ValueError("private key not present")
        # We need to do lots of if-not-this-then-start-over type tests.
        # A while loop with continue statements is the cleanest way to do so.
        while True:
            k = int(hexlify(os.urandom(self.BITLENGTH // 8)), 16) % q
            if k == 0:
                continue
            r = pow(g, k, p) % q
            if r == 0:
                continue
            h = (int(self.HASHMOD(data).hexdigest(), 16) + (x * r)) % q
            s = (modinv(k, q) * h) % q
            if s == 0:
                continue
            break
        assert 0 < r < q
        assert 0 < s < q
        bytelength = self.BITLENGTH // 8
        r_bytes = int2bytes(r).rjust(bytelength, b"\x00")
        s_bytes = int2bytes(s).rjust(bytelength, b"\x00")
        return r_bytes + s_bytes


def modinv(a, m):
    """Find the modular inverse of a, with modulus m."""
    # This is a transliteration of the algorithm as it was described
    # to me by Wikipedia, using the Extended Euclidean Algorithm.
    x = 0
    lastx = 1
    y = 1
    lasty = 0
    b = m
    while b != 0:
        a, (q, b) = b, divmod(a, b)
        x, lastx = lastx - (q * x), x
        y, lasty = lasty - (q * y), y
    return lastx % m


def int2bytes(x):
    """Convert a Python integer to bigendian bytestring."""
    # It's faster to go via hex encoding in C code than it is to try
    # encoding directly into binary with a python-level loop.
    # (and hex-slice-strip seems consistently faster than using "%x" format)
    hexbytes = hex(x)[2:].rstrip("L").encode("ascii")
    if len(hexbytes) % 2:
        hexbytes = b"0" + hexbytes
    return unhexlify(hexbytes)


def _check_keys(data, keys):
    """Verify that the given data dict contains the specified keys."""
    for key in keys:
        if not key in data:
            msg = 'missing %s in data - %s' % (key, str(data.keys()))
            raise ValueError(msg)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""

Crypto primitives built on top of M2Crypto.

This file provides the "fast path" crypto implementation for PyBrowserID.
It uses the public-key-crypto routines from M2Crypto for nice fast operation.

There is also a pure-python fallback module that's slower, but avoid
having to install M2Crypto.

"""

import struct
from binascii import hexlify, unhexlify

from M2Crypto import BIO

from browserid.crypto._m2_monkeypatch import m2
from browserid.crypto._m2_monkeypatch import DSA as _DSA
from browserid.crypto._m2_monkeypatch import RSA as _RSA


class Key(object):
    """Generic base class for Key objects."""

    KEY_MODULE = None

    @classmethod
    def from_pem_data(cls, data=None, filename=None):
        """Alternative constructor for loading from PEM format data."""
        self = cls.__new__(cls)
        if data is not None:
            bio = BIO.MemoryBuffer(str(data))
        elif filename is not None:
            bio = BIO.openfile(filename)
        else:
            msg = "Please specify either 'data' or 'filename' argument."
            raise ValueError(msg)
        self.keyobj = self.KEY_MODULE.load_pub_key_bio(bio)
        return self

    def to_pem_data(self):
        """Save the public key data to a PEM format string."""
        b = BIO.MemoryBuffer()
        try:
            self.keyobj.save_pub_key_bio(b)
            return b.getvalue()
        finally:
            b.close()

    def verify(self, signed_data, signature):
        """Verify the given signature."""
        raise NotImplementedError  # pragma: nocover

    def sign(self, data):
        """Sign the given data."""
        raise NotImplementedError  # pragma: nocover


#
#  RSA keys, implemented using the RSA support in M2Crypto.
#

class RSKey(Key):

    KEY_MODULE = _RSA
    DIGESTSIZE = None
    HASHNAME = None
    HASHMOD = None

    def __init__(self, data):
        _check_keys(data, ('e', 'n'))
        e = int2mpint(int(data["e"]))
        n = int2mpint(int(data["n"]))
        try:
            d = int2mpint(int(data["d"]))
        except KeyError:
            self.keyobj = _RSA.new_pub_key((e, n))
        else:
            self.keyobj = _RSA.new_key((e, n, d))

    def verify(self, signed_data, signature):
        digest = self.HASHMOD(signed_data).digest()
        try:
            return self.keyobj.verify(digest, signature, self.HASHNAME)
        except _RSA.RSAError:
            return False

    def sign(self, data):
        digest = self.HASHMOD(data).digest()
        return self.keyobj.sign(digest, self.HASHNAME)


#
#  DSA keys, implemented using the DSA support in M2Crypto, along with
#  some formatting tweaks to match what the browserid node-js server does.
#

class DSKey(Key):

    KEY_MODULE = _DSA
    BITLENGTH = None
    HASHMOD = None

    def __init__(self, data):
        _check_keys(data, ('p', 'q', 'g', 'y'))
        self.p = p = long(data["p"], 16)
        self.q = q = long(data["q"], 16)
        self.g = g = long(data["g"], 16)
        self.y = y = long(data["y"], 16)
        if "x" not in data:
            self.x = None
            self.keyobj = _DSA.load_pub_key_params(int2mpint(p), int2mpint(q),
                                                   int2mpint(g), int2mpint(y))
        else:
            self.x = x = long(data["x"], 16)
            self.keyobj = _DSA.load_key_params(int2mpint(p), int2mpint(q),
                                               int2mpint(g), int2mpint(y),
                                               int2mpint(x))

    @classmethod
    def from_pem_data(cls, data=None, filename=None):
        self = super(DSKey, cls).from_pem_data(data, filename)
        self.p = mpint2int(m2.dsa_get_p(self.keyobj.dsa))
        self.q = mpint2int(m2.dsa_get_q(self.keyobj.dsa))
        self.g = mpint2int(m2.dsa_get_g(self.keyobj.dsa))
        self.y = None
        self.x = None
        return self

    def verify(self, signed_data, signature):
        # Restore any leading zero bytes that might have been stripped.
        signature = hexlify(signature)
        hexlength = self.BITLENGTH // 4
        signature = signature.rjust(hexlength * 2, b"0")
        if len(signature) != hexlength * 2:
            return False
        # Split the signature into "r" and "s" components.
        r = long(signature[:hexlength], 16)
        s = long(signature[hexlength:], 16)
        if r <= 0 or r >= self.q:
            return False
        if s <= 0 or s >= self.q:
            return False
        # Now we can check the digest.
        digest = self.HASHMOD(signed_data).digest()
        return self.keyobj.verify(digest, int2mpint(r), int2mpint(s))

    def sign(self, data):
        if not self.x:
            raise ValueError("private key not present")
        digest = self.HASHMOD(data).digest()
        r, s = self.keyobj.sign(digest)
        # We need precisely "bytelength" bytes from each integer.
        # M2Crypto might give us more or less, so snip and pad appropriately.
        bytelength = self.BITLENGTH // 8
        r_bytes = r[4:].rjust(bytelength, b"\x00")[-bytelength:]
        s_bytes = s[4:].rjust(bytelength, b"\x00")[-bytelength:]
        return r_bytes + s_bytes


#
#  Other helper functions.
#


def int2mpint(x):
    """Convert a Python long integer to a string in OpenSSL's MPINT format."""
    # MPINT is big-endian bytes with a size prefix.
    # It's faster to go via hex encoding in C code than it is to try
    # encoding directly into binary with a python-level loop.
    # (and hex-slice-strip seems consistently faster than using "%x" format)
    hexbytes = hex(x)[2:].rstrip("L").encode("ascii")
    if len(hexbytes) % 2:
        hexbytes = b"0" + hexbytes
    bytes = unhexlify(hexbytes)
    # Add an extra significant byte that's just zero.  I think this is only
    # necessary if the number has its MSB set, to prevent it being mistaken
    # for a sign bit.  I do it uniformly since it's valid and simpler.
    return struct.pack(">I", len(bytes) + 1) + b"\x00" + bytes


def mpint2int(data):
    """Convert a string in OpenSSL's MPINT format to a Python long integer."""
    hexbytes = hexlify(data[4:])
    return long(hexbytes, 16)


def _check_keys(data, keys):
    """Verify that the given data dict contains the specified keys."""
    for key in keys:
        if not key in data:
            msg = 'missing %s in data - %s' % (key, str(data.keys()))
            raise ValueError(msg)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import json

from browserid import netutils
from browserid.verifiers import Verifier
from browserid.errors import (InvalidSignatureError,
                              ConnectionError,
                              AudienceMismatchError)

BROWSERID_VERIFIER_URL = "https://verifier.login.persona.org/verify"


class RemoteVerifier(Verifier):
    """Class for remote verification of BrowserID identity assertions.

    This class submits assertions to the persona.org verifier service
    for remote verification.  It's slower but potentially a little bit
    safer than the still-under-development LocalVerifier class.
    """

    def __init__(self, audiences=None, verifier_url=None):
        if verifier_url is None:
            verifier_url = BROWSERID_VERIFIER_URL
        super(RemoteVerifier, self).__init__(audiences)
        self.verifier_url = verifier_url

    def verify(self, assertion, audience=None):
        """Verify the given BrowserID assertion.

        This method posts the given BrowserID assertion to the remote verifier
        service.  If it is successfully verified then a dict giving the
        email and audience is returned.  If it is not valid then an error
        is raised.

        If the 'audience' argument is given, it first verifies that the
        audience of the assertion matches the one given.  This can help
        avoid doing lots of crypto for assertions that can't be valid.
        If you don't specify an audience, you *MUST* validate the audience
        value returned by this method.
        """
        # Check the audience locally.
        # No point talking to the network if we know it's going to fail.
        # If no explicit audience was given, this will also parse it out
        # for inclusion in the request to the remote verifier service.
        audience = self.check_audience(assertion, audience)

        response = netutils.post(self.verifier_url, {'assertion': assertion,
                                                     'audience': audience})

        # BrowserID server sends "500 server error" for broken assertions.
        # For now, just translate that directly.  Should check by hand.
        if response.status_code == 500:
            raise ValueError('Malformed assertion')

        try:
            data = json.loads(response.text)
        except ValueError:
            raise ConnectionError("server returned invalid response")

        # Did it come back clean?
        if data.get('status') != "okay":
            raise InvalidSignatureError(str(data))
        if data.get('audience') != audience:
            raise AudienceMismatchError(data.get("audience"), audience)
        return data



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import threading
import multiprocessing

from browserid.verifiers.local import LocalVerifier


class WorkerPoolVerifier(object):
    """Class for verifying against a pool of worker processes.

    This class is a drop-in replacement for LocalVerifier that uses a pool
    of worker processes to side-step the GIL.  It can be very helpful on
    systems with lots of cores.

    By default a WorkerPoolVerifier will spawn as many processes as you
    have cpus on your system, and will check assertions against an instance
    of LocalVerifier.  You can customize this with the arguments "num_procs"
    and "verifier" respectively.
    """

    def __init__(self, num_procs=None, verifier=None):
        # Try to choose a sensible number of processes by default.
        if num_procs is None:
            try:
                num_procs = multiprocessing.cpu_count()
            except NotImplementedError:
                num_procs = 2
        # Use LocalVerifier by default, but allow overriding.
        if verifier is None:
            verifier = LocalVerifier()
        self.num_procs = num_procs
        self.verifier = verifier
        # Create the various communication channels.
        # Yes, this duplicates a lot of the logic from multprocessing.Pool.
        # I don't want to have to constantly pickle the verifier object
        # to send it into the subprocesses, and the Pool class doesn't lend
        # itself to subclases in a way that would avoid this.  So here we are.
        # We have:
        # 1) a queue from which the workers read jobs
        self._work_queue = multiprocessing.Queue()
        # 2) a queue into which the workers push results
        self._result_queue = multiprocessing.Queue()
        # 3) a thread that dispatches results to other waiting threads,
        #    by signalling on a condition object.
        self._lock = threading.Lock()
        self._waiting_conds = {}
        self._spare_conds = []
        # Now we can start the required processes and threads.
        self._result_thread = threading.Thread(target=self._run_result_thread)
        self._result_thread.start()
        self._processes = []
        for n in range(num_procs):
            proc = multiprocessing.Process(target=self._run_worker)
            self._processes.append(proc)
            proc.start()

    def __del__(self):
        self.close()

    def verify(self, *args, **kwds):
        """Verify the given BrowserID assertion.

        This method parses a BrowserID identity assertion, verifies the bundled
        chain of certificates and signatures, and returns the extracted
        email address and audience.

        If the 'audience' argument is given, it first verifies that the
        audience of the assertion matches the one given.  This can help
        avoid doing lots of crypto for assertions that can't be valid.
        If you don't specify an audience, you *MUST* validate the audience
        value returned by this method.

        Any other positional or keyword arguments will be passed on to the
        underlying verifier object.
        """
        with self._lock:
            # Get a condition with which to wait for the result.
            try:
                cond = self._spare_conds.pop()
            except IndexError:
                cond = threading.Condition(self._lock)
            job_id = id(cond)
            self._waiting_conds[job_id] = cond
            # Send the job to the workers, wait for the result.
            self._work_queue.put((job_id, args, kwds))
            cond.wait()
            ok, result = self._waiting_conds.pop(job_id)
            # Store the condition for reuse.
            self._spare_conds.append(cond)
            if ok:
                return result
            else:
                raise result

    def close(self):
        """Close down the worker processes.

        This method shuts down the various background processes and threads
        that are uses by the WorkerPoolVerifier.  It's good hygiene to call
        this explicitly, but it will be called by the destructor if you forget.
        """
        with self._lock:
            if self._work_queue is None:
                return
            for x in range(self.num_procs):
                self._work_queue.put((None, None, None))
            for proc in self._processes:
                proc.join()
            self._result_queue.put((None, None, None))
            self._result_thread.join()
            self._work_queue = None

    def _run_result_thread(self):
        """Method to run for the background result-dispatching thread.

        This method loops through results that are returned by the workers,
        dispatching them to the appropraite waiting thread.
        """
        while True:
            job_id, ok, result = self._result_queue.get()
            if job_id is None:
                break
            with self._lock:
                cond = self._waiting_conds[job_id]
                self._waiting_conds[job_id] = (ok, result)
                cond.notify()

    def _run_worker(self):
        """Method to run for the background worker processes.

        This method loops through jobs from the work queue, executing them
        with the verifier object and pushing the result back via the result
        queue.
        """
        while True:
            job_id, args, kwds = self._work_queue.get()
            if job_id is None:
                break
            try:
                result = self.verifier.verify(*args, **kwds)
                ok = True
            except Exception as e:
                result = e
                ok = False
            self._result_queue.put((job_id, ok, result))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import time
import warnings

from unittest.mock import Mock, patch

import browserid
from browserid.tests.support import (patched_supportdoc_fetching,
                                     get_keypair,
                                     make_assertion, unittest, callwith)
from browserid import jwt
from browserid import RemoteVerifier, LocalVerifier
from browserid.supportdoc import FIFOCache, SupportDocumentManager
from browserid.verifiers.workerpool import WorkerPoolVerifier
from browserid.utils import (encode_json_bytes,
                             decode_json_bytes,
                             bundle_certs_and_assertion)
from browserid.errors import (TrustError,
                              ConnectionError,
                              ExpiredSignatureError,
                              InvalidSignatureError,
                              AudienceMismatchError)

# This is an old assertion I generated on myfavoritebeer.org.
# It's expired and signed with an old private key.
EXPIRED_ASSERTION = """
    eyJjZXJ0aWZpY2F0ZXMiOlsiZXlKaGJHY2lPaUpTVXpJMU5pSjkuZXlKcGMzTWlPaUppY2
    05M2MyVnlhV1F1YjNKbklpd2laWGh3SWpveE16SXpPVGcyTURJM01qUXdMQ0pwWVhRaU9q
    RXpNak00T1RrMk1qY3lOREFzSW5CMVlteHBZeTFyWlhraU9uc2lZV3huYjNKcGRHaHRJam
    9pUkZNaUxDSjVJam9pT1RCa056azBabVkzWmpVeE56WTNZVEJpWm1GaU5UQTVOVFZrTXpR
    ek1qQmxNemhtTldKbE1qSmhaVFE1TURNME5qQmlabUk0TmpFMk1qTTVOamhtTW1Oa09EQX
    lOemd4Tm1FeU5ESXhaR0ZqWWpnNU9UTTJOalF6WWpRNVlXRTVOMkU1T0RSa01UVmpaakU0
    WXpZeFpEUXlNemcyWWpRM01UQTJNR0kyWlRVMU9HSXhObVZsTTJZMVkyWmlaVEZoTmpVeU
    5tUXpNelpoT0RFMVlUZzROelExTXpFNU1UVmpOVEk1TlRoaVpqWXdaalV6TkRoaU5XTXdO
    RGc0T0Raa00yRTNNVGt5WWpBeU9EWm1ZV0kzT1Roak5HVmtNekkzWlRGa01HUXlZMk0yWk
    RKa05tTmpOV014T0dObU5UWXhORFZqTm1Zd01qVXpNamswWldZeU0yVmhZbVk0TXpSa05E
    aGlOakZoTldRME9UUm1PREptTWpWak5XUmlZbUV6TVRreE16UmtZbU13WlRka05EbGhaRE
    kyWW1NeE1EWXdaR016TW1GaU5tTTNabUpoWkdNNVpHRmtaV0ZpWlRZMllXUmpaR0poTm1F
    ME5ESm1PV1poWmprelkyUXdOV000WVRFNU5ERTFaalF5WldFMk9HRXlNV1ZtWlRWbE16UX
    hNV1poWm1Wa05XVm1NR1ZqTjJVd01XRmlNMlExTXpsbE16TmxOemMxWmpWbU0yVXdZVFpr
    TkRnNU16WmtOVEprT0RNeU5ERm1ZbU5rWW1abFpXTXdZalJrWmprME5EWTRNbU0wWVRZMk
    4ySmpOekJsTm1NME5UQTBObUptWlRrek9UUmlNMkkwTW1NeU5UQXdOMlF5WXpJMVpUaGpO
    MlZoWWpObFpETXhZVGNpTENKd0lqb2laRFpqTkdVMU1EUTFOamszTnpVMll6ZGhNekV5Wk
    RBeVl6SXlPRGxqTWpWa05EQm1PVGsxTkRJMk1XWTNZalU0TnpZeU1UUmlObVJtTVRBNVl6
    Y3pPR0kzTmpJeU5tSXhPVGxpWWpkbE16Tm1PR1pqTjJGak1XUmpZek14Tm1VeFpUZGpOem
    c1TnpNNU5URmlabU0yWm1ZeVpUQXdZMk01T0RkalpEYzJabU5tWWpCaU9HTXdNRGsyWWpC
    aU5EWXdabVptWVdNNU5qQmpZVFF4TXpaak1qaG1OR0ptWWpVNE1HUmxORGRqWmpkbE56a3
    pOR016T1RnMVpUTmlNMlE1TkROaU56ZG1NRFpsWmpKaFpqTmhZek0wT1RSbVl6TmpObVpq
    TkRrNE1UQmhOak00TlRNNE5qSmhNREppWWpGak9ESTBZVEF4WWpkbVl6WTRPR1UwTURJNE
    5USTNZVFU0WVdRMU9HTTVaRFV4TWpreU1qWTJNR1JpTldRMU1EVmlZekkyTTJGbU1qa3pZ
    bU01TTJKalpEWmtPRGcxWVRFMU56VTNPV1EzWmpVeU9UVXlNak0yWkdRNVpEQTJZVFJtWX
    pOaVl6SXlORGRrTWpGbU1XRTNNR1kxT0RRNFpXSXdNVGMyTlRFek5UTTNZems0TTJZMVlU
    TTJOek0zWmpBeFpqZ3lZalEwTlRRMlpUaGxOMll3Wm1GaVl6UTFOMlV6WkdVeFpEbGpOV1
    JpWVRrMk9UWTFZakV3WVRKaE1EVTRNR0l3WVdRd1pqZzRNVGM1WlRFd01EWTJNVEEzWm1J
    M05ETXhOR0V3TjJVMk56UTFPRFl6WW1NM09UZGlOekF3TW1WaVpXTXdZakF3TUdFNU9HVm
    lOamszTkRFME56QTVZV014TjJJME1ERWlMQ0p4SWpvaVlqRmxNemN3WmpZME56SmpPRGMx
    TkdOalpEYzFaVGs1TmpZMlpXTTRaV1l4Wm1RM05EaGlOelE0WW1KaVl6QTROVEF6WkRneV
    kyVTRNRFUxWVdJellpSXNJbWNpT2lJNVlUZ3lOamxoWWpKbE0ySTNNek5oTlRJME1qRTNP
    V1E0Wmpoa1pHSXhOMlptT1RNeU9UZGtPV1ZoWWpBd016YzJaR0l5TVRGaE1qSmlNVGxqT0
    RVMFpHWmhPREF4Tmpaa1pqSXhNekpqWW1NMU1XWmlNakkwWWpBNU1EUmhZbUl5TW1SaE1t
    TTNZamM0TlRCbU56Z3lNVEkwWTJJMU56VmlNVEUyWmpReFpXRTNZelJtWXpjMVlqRmtOem
    MxTWpVeU1EUmpaRGRqTWpOaE1UVTVPVGt3TURSak1qTmpaR1ZpTnpJek5UbGxaVGMwWlRn
    NE5tRXhaR1JsTnpnMU5XRmxNRFZtWlRnME56UTBOMlF3WVRZNE1EVTVNREF5WXpNNE1UbG
    hOelZrWXpka1kySmlNekJsTXpsbFptRmpNelpsTURkbE1tTTBNRFJpTjJOaE9UaGlNall6
    WWpJMVptRXpNVFJpWVRrell6QTJNalUzTVRoaVpEUTRPV05sWVRaa01EUmlZVFJpTUdJM1
    pqRTFObVZsWWpSak5UWmpORFJpTlRCbE5HWmlOV0pqWlRsa04yRmxNR1ExTldJek56a3lN
    alZtWldJd01qRTBZVEEwWW1Wa056Sm1Nek5sTURZMk5HUXlPVEJsTjJNNE5EQmtaak5sTW
    1GaVlqVmxORGd4T0RsbVlUUmxPVEEyTkRabU1UZzJOMlJpTWpnNVl6WTFOakEwTnpZM09U
    bG1OMkpsT0RReU1HRTJaR013TVdRd056aGtaVFF6TjJZeU9EQm1abVl5WkRka1pHWXhNal
    E0WkRVMlpURmhOVFJpT1RNellUUXhOakk1WkRaak1qVXlPVGd6WXpVNE56azFNVEExT0RB
    eVpETXdaRGRpWTJRNE1UbGpaalpsWmlKOUxDSndjbWx1WTJsd1lXd2lPbnNpWlcxaGFXd2
    lPaUp5ZVdGdVFISm1heTVwWkM1aGRTSjlmUS5rRTQzY0NrQ2d0Z1J5TUE5N1c4Rmo2a3hG
    bUhVMGdUdVNlelhFWWxldktPZGVxNFhJV0RqeHBWUy1Fekt3X0s1bTkyS2M3dXlHSy1nRl
    VvaENjc2gxSFhHNXhBOXRwMzJvOXhFelFrWEVOR1pPU3VZMUN4NGNmOWNiYkh4UkdLdGVR
    S1RXVUdZVEhLMWJRZ09hMEFNaGpZMmc3eUwtbk5SMGJES2dBMDE0b3VkSjhjMVVQYm10dG
    1FQjRoZk43aEVVeVZKb0hDSVdGeTV1TlpiV2Q4X1NtSXhvUk9TS2dzZzNrdDJ6bWRiaWdE
    Yks3ZmFKZkMtUEg0ZVpSM1Q2dkZWTDhkTk4yMzhCN24yWlp6N1kzU1BtZ3Y3QmlGNGRJYl
    NkQmZjb2dGMlhsZHBVNTRRNE5xeTBSQjg4TUV1eWNWajZObmhyUThWOFVYRFdEQ21TaUJt
    bmciXSwiYXNzZXJ0aW9uIjoiZXlKaGJHY2lPaUpFVXpJMU5pSjkuZXlKbGVIQWlPakV6TW
    pNNE9UazNORGMyTURVc0ltRjFaQ0k2SW1oMGRIQTZMeTl0ZVdaaGRtOXlhWFJsWW1WbGNp
    NXZjbWNpZlEuZDRITjc5WnBFR0x1blVBbnNBcjFKRXAyTml0djUzTy1ib1BGNnZ0RzA5QV
    U0MGdaNzRkTi1FTTV5TnBINDZLcUpRTXZKbzlHeUhoT1hoekZZT1R2Z1EifQ
""".replace(" ", "").replace("\n", "").strip()


class VerifierTestCases(object):
    """Generic testcases for Verifier implementations."""

    def test_expired_assertion(self):
        self.assertRaises(TrustError, self.verifier.verify, EXPIRED_ASSERTION)

    def test_junk(self):
        self.assertRaises(ValueError, self.verifier.verify, "JUNK")
        self.assertRaises(ValueError, self.verifier.verify, "J")
        self.assertRaises(ValueError, self.verifier.verify, "\x01\x02")

    def test_malformed_assertions(self):
        errors = (ValueError, TrustError)
        # This one doesn't actually contain an assertion
        assertion = encode_json_bytes({})
        self.assertRaises(errors, self.verifier.verify, assertion)
        # This one has no certificates
        pub, priv = get_keypair("TEST")
        assertion = bundle_certs_and_assertion(
            [],
            jwt.generate({"aud": "TEST"}, priv),
        )
        self.assertRaises(errors, self.verifier.verify, assertion)
        # This one has too many certificates in the chain.
        assertion = bundle_certs_and_assertion(
            [jwt.generate({}, priv), jwt.generate({}, priv)],
            jwt.generate({"aud": "TEST"}, priv),
        )
        self.assertRaises(errors, self.verifier.verify, assertion)


class TestLocalVerifier(unittest.TestCase, VerifierTestCases):

    def setUp(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            self.verifier = LocalVerifier(["*"])
        # There should be no warning about using this verifier.
        self.assertEquals(len(w), 0)

    @callwith(patched_supportdoc_fetching())
    def test_error_handling_in_verify_certificate_chain(self):
        self.assertRaises(ValueError,
                          self.verifier.verify_certificate_chain, [])
        certs = decode_json_bytes(EXPIRED_ASSERTION)["certificates"]
        certs = [jwt.parse(cert) for cert in certs]
        self.assertRaises(ExpiredSignatureError,
                          self.verifier.verify_certificate_chain, certs)

    @callwith(patched_supportdoc_fetching())
    def test_well_known_doc_with_public_key(self):
        assertion = make_assertion("t@m.com", "http://e.com")
        self.assertTrue(self.verifier.verify(assertion))

    @callwith(patched_supportdoc_fetching())
    def test_delegated_primary(self):
        assertion = make_assertion("t@redirect.org", "http://persona.org",
                issuer="delegated.org")
        self.assertTrue(self.verifier.verify(assertion))

    @callwith(patched_supportdoc_fetching())
    def test_double_delegated_primary(self):
        assertion = make_assertion("t@redirect-twice.org",
                "http://persona.org", issuer="delegated.org")
        self.assertTrue(self.verifier.verify(assertion))

    @callwith(patched_supportdoc_fetching())
    def test_extraction_of_extra_claims(self):
        assertion = make_assertion("t@m.com", "http://e.com",
            idp_claims={"extra": "claim"},
            user_claims={"another": "claim"},
        )
        result = self.verifier.verify(assertion)
        self.assertEqual(result["status"], "okay")
        self.assertEqual(result["idpClaims"], {"extra": "claim"})
        self.assertEqual(result["userClaims"], {"another": "claim"})

    @callwith(patched_supportdoc_fetching())
    def test_extraction_of_extra_claims_from_principal(self):
        assertion = make_assertion("t@m.com", "http://e.com",
            idp_claims={
                "extra": "claim",
                "principal": {
                    "email": "t@m.com",
                    "another": "claim",
                    "public-key": "should-be-ignored",
                }
            },
        )
        result = self.verifier.verify(assertion)
        self.assertEqual(result["status"], "okay")
        self.assertEqual(result["idpClaims"], {
            "extra": "claim",
            "another": "claim",
        })
        self.assertTrue("userClaims" not in result)


    @callwith(patched_supportdoc_fetching())
    def test_email_validation(self):
        verifier = LocalVerifier(audiences="http://persona.org")

        # Null bytes in the email hostname.
        assertion = make_assertion("test@users.example\x00.com",
                                   "http://persona.org")
        self.assertRaises(ValueError, verifier.verify, assertion)

        # Newlines in the email hostanem.
        assertion = make_assertion("test@users.example.com\n@example.com",
                                   "http://persona.org")
        self.assertRaises(ValueError, verifier.verify, assertion)

        # Null bytes in the email username.
        assertion = make_assertion(u"test\u0000@users.example.com",
                                   "http://persona.org")
        self.assertRaises(ValueError, verifier.verify, assertion)

        # Null bytes with regex-busting newline.
        assertion = make_assertion(u"test@example.com\u0000\n@evil.com",
                                   "http://persona.org")
        self.assertRaises(ValueError, verifier.verify, assertion)

    @callwith(patched_supportdoc_fetching())
    def test_audience_verification(self):

        # create an assertion with the audience set to http://persona.org for
        # the tests. This assertion is only valid for this audience.
        assertion = make_assertion("alexis@mozilla.com", "http://persona.org")

        # we don't set any audience explicitely here
        verifier = LocalVerifier()

        # specifying the audience on verifier.verify uses it.
        self.assertRaises(AudienceMismatchError, verifier.verify, assertion,
                          audience="*.example.com")

        # if we change the audience to the expected one, the assertion is
        # considered valid
        self.assertTrue(verifier.verify(assertion, audience="persona.org"))

        # specifying the audience when creating the verifier AND when calling
        # verifier.verify.
        verifier = LocalVerifier(["*.example.com"])
        self.assertRaises(AudienceMismatchError, verifier.verify, assertion,
                          audience="*.example.com")

        # specifying a different audience at instanciation and at verification,
        # only the last one is used.
        self.assertTrue(verifier.verify(assertion, audience="persona.org"))

        # overwritting the audience with an invalid one (we are waiting for
        # persona.org but getting example.com) raises an error
        self.assertRaises(AudienceMismatchError, verifier.verify,
                          audience="persona.org",
                          assertion=make_assertion("alexis@mozilla.com",
                                                   "http://example.com"))

        # the assertion is valid for http://persona.org; the verifier is
        # configured to accept this audience so it should validate
        verifier = LocalVerifier(["persona.org"])
        self.assertTrue(verifier.verify(assertion))

        # but if we ask explicitely for a different audience (the assertion is
        # not accepted, even if the instance is configured so)
        self.assertRaises(AudienceMismatchError, verifier.verify,
                          assertion, audience="example.com")


class TestRemoteVerifier(unittest.TestCase, VerifierTestCases):

    def setUp(self):
        self.verifier = RemoteVerifier(["*"])

    @patch('browserid.netutils.requests')
    def _verify(self, requests, response_text='', assertion=EXPIRED_ASSERTION,
                status_code=200):
        response = Mock()
        response.text = response_text
        response.status_code = status_code
        requests.request.return_value = response

        return self.verifier.verify(assertion)

    def test_expired_assertion(self):
        try:
            return super(TestRemoteVerifier, self).test_expired_assertion()
        except ConnectionError:
            raise unittest.SkipTest("error connecting to remote server")

    def test_malformed_assertions(self):
        try:
            return super(TestRemoteVerifier, self).test_malformed_assertions()
        except ConnectionError:
            raise unittest.SkipTest("error connecting to remote server")

    def test_handling_of_valid_response_from_server(self):
        response_text = ('{"email": "t@m.com", "status": "okay", '
                         '"audience": "http://myfavoritebeer.org"}')
        data = self._verify(response_text=response_text)
        self.assertEquals(data["email"], "t@m.com")

    def test_handling_of_invalid_json_from_server(self):
        with self.assertRaises(ConnectionError):
            self._verify(response_text='SERVER RETURNS INVALID JSON')

    @patch('browserid.netutils.requests')
    def test_handling_of_incorrect_audience_returned_by_server(self, requests):
        response_text = ('{"email": "t@m.com", "status": "okay", '
                         '"audience": "WRONG"}')
        with self.assertRaises(AudienceMismatchError):
            self._verify(response_text=response_text)

    @patch('browserid.netutils.requests')
    def test_handling_of_500_error_from_server(self, requests):
        with self.assertRaises(ValueError):
            self._verify(status_code=500)

    def test_handling_of_503_error_from_server(self):
        with self.assertRaises(ConnectionError):
            self._verify(status_code=503)


class TestDummyVerifier(unittest.TestCase, VerifierTestCases):

    def setUp(self):
        self.patched = patched_supportdoc_fetching()
        self.patched.__enter__()
        self.verifier = LocalVerifier(["*"])

    def tearDown(self):
        self.patched.__exit__(None, None, None)

    def test_verification_of_valid_dummy_assertion(self):
        audience = "http://example.com"
        assertion = make_assertion("test@example.com", audience)
        self.assertTrue(self.verifier.verify(assertion))
        self.assertTrue(self.verifier.verify(assertion, audience))
        self.assertRaises(AudienceMismatchError,
                          self.verifier.verify, assertion, "http://moz.com")

    def test_verification_of_oldstyle_dummy_assertion(self):
        audience = "http://example.com"
        assertion = make_assertion("test@example.com", audience,
                                   new_style=False)
        self.assertTrue(self.verifier.verify(assertion))
        self.assertTrue(self.verifier.verify(assertion, audience))
        self.assertRaises(AudienceMismatchError, self.verifier.verify,
                          assertion, "http://moz.com")

    def test_verification_of_untrusted_issuer(self):
        audience = "http://example.com"
        issuer = "moz.com"
        # Assertions for @moz.com addresses can come from moz.com
        assertion = make_assertion("test@moz.com", audience, issuer=issuer)
        self.assertTrue(self.verifier.verify(assertion, audience))
        # But assertions for other addresses cannot (unless they delegated).
        assertion = make_assertion("test@example.com", audience,
                                   issuer=issuer)
        self.assertRaises(InvalidSignatureError, self.verifier.verify,
                          assertion, audience)

    def test_verification_of_expired_dummy_assertion(self):
        audience = "http://example.com"
        now = (time.time() * 1000)
        assertion = make_assertion("test@example.com", audience, exp=now - 1)
        self.assertTrue(self.verifier.verify(assertion, now=now - 2))
        self.assertRaises(ExpiredSignatureError, self.verifier.verify,
                          assertion)

    def test_verification_of_dummy_assertion_with_bad_assertion_sig(self):
        audience = "http://example.com"
        assertion = make_assertion("test@example.com", audience,
                                   assertion_sig="BADTOTHEBONE")
        self.assertRaises(InvalidSignatureError, self.verifier.verify,
                          assertion)

    def test_verification_of_dummy_assertion_with_bad_certificate_sig(self):
        audience = "http://example.com"
        assertion = make_assertion("test@example.com", audience,
                                   certificate_sig="CORRUPTUS")
        self.assertRaises(InvalidSignatureError, self.verifier.verify,
                          assertion)

    def test_cache_eviction_based_on_time(self):
        supportdocs = SupportDocumentManager(FIFOCache(cache_timeout=0.1))
        verifier = LocalVerifier(["*"], supportdocs=supportdocs)
        # Prime the cache by verifying an assertion.
        assertion = make_assertion("test@example.com", "")
        self.assertTrue(verifier.verify(assertion))
        # Make it error out if re-fetching the keys

        exc = RuntimeError("key fetch disabled")
        with patched_supportdoc_fetching(exc=exc):
            # It should be in the cache, so this works fine.
            self.assertTrue(verifier.verify(assertion))
            # But after sleeping it gets evicted and the error is triggered.
            time.sleep(0.1)
            self.assertRaises(RuntimeError, verifier.verify, assertion)

    def test_cache_eviction_based_on_size(self):
        supportdocs = SupportDocumentManager(max_size=2)
        verifier = LocalVerifier(["*"], supportdocs=supportdocs)
        # Prime the cache by verifying some assertions.
        assertion1 = make_assertion("test@1.com", "", "1.com")
        self.assertTrue(verifier.verify(assertion1))
        assertion2 = make_assertion("test@2.com", "", "2.com")
        self.assertTrue(verifier.verify(assertion2))
        self.assertEquals(len(supportdocs.cache), 2)
        # Hitting a third host should evict the first public key.
        assertion3 = make_assertion("test@3.com", "", "3.com")
        self.assertTrue(verifier.verify(assertion3))
        self.assertEquals(len(supportdocs.cache), 2)
        # Make it error out if re-fetching any keys

        exc = RuntimeError("key fetch disabled")
        with patched_supportdoc_fetching(exc=exc):
            # It should have to re-fetch for 1, but not 2.
            self.assertTrue(verifier.verify(assertion2))
            self.assertRaises(RuntimeError, verifier.verify, assertion1)

    def test_cache_eviction_during_write(self):
        supportdocs = SupportDocumentManager(cache_timeout=0.1)
        verifier = LocalVerifier(["*"], supportdocs=supportdocs)
        # Prime the cache by verifying an assertion.
        assertion1 = make_assertion("test@1.com", "", "1.com")
        self.assertTrue(verifier.verify(assertion1))
        self.assertEquals(len(supportdocs.cache), 1)
        # Let that cached key expire
        time.sleep(0.1)
        # Now grab a different key; caching it should purge the expired key.
        assertion2 = make_assertion("test@2.com", "", "2.com")
        self.assertTrue(verifier.verify(assertion2))
        self.assertEquals(len(supportdocs.cache), 1)
        # Check that only the second entry is in cache.

        exc = RuntimeError("key fetch disabled")
        with patched_supportdoc_fetching(exc=exc):
            self.assertTrue(verifier.verify(assertion2))
            self.assertRaises(RuntimeError, verifier.verify, assertion1)

    def test_audience_pattern_checking(self):
        verifier = LocalVerifier(["*.moz.com", "www.test.com"])
        # Domains like *.moz.com should be valid audiences.
        # They will work with both the implicit patterns and explicit audience.
        assertion = make_assertion("test@example.com", "www.moz.com")
        self.assertTrue(verifier.verify(assertion))
        self.assertTrue(verifier.verify(assertion, "www.moz.com"))
        self.assertRaises(AudienceMismatchError,
                          verifier.verify, assertion, "www.test.com")
        # The specific domain www.test.com should be a valid audience.
        # It will work with both the implicit patterns and explicit audience.
        assertion = make_assertion("test@example.com", "www.test.com")
        self.assertTrue(verifier.verify(assertion))
        self.assertTrue(verifier.verify(assertion, "www.test.com"))
        self.assertTrue(verifier.verify(assertion, "*.test.com"))
        self.assertRaises(AudienceMismatchError,
                          verifier.verify, assertion, "www.moz.com")
        # Domains not matching any patterns should not be valid audiences.
        # They will fail unless given as an explicit argument.
        assertion = make_assertion("test@example.com", "www.evil.com")
        self.assertRaises(AudienceMismatchError, verifier.verify, assertion)
        self.assertTrue(verifier.verify(assertion, "www.evil.com"))
        self.assertTrue(verifier.verify(assertion, "*.evil.com"))


class TestWorkerPoolVerifier(TestDummyVerifier):

    def setUp(self):
        super(TestWorkerPoolVerifier, self).setUp()
        self.verifier = WorkerPoolVerifier(
                verifier=LocalVerifier(["*"])
        )

    def tearDown(self):
        super(TestWorkerPoolVerifier, self).tearDown()
        self.verifier.close()


class TestShortcutFunction(unittest.TestCase):

    def test_shortcut(self):
        self.assertRaises(TrustError, browserid.verify, EXPIRED_ASSERTION)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""

Python library for the BrowserID identity protocol.

"""

__ver_major__ = 0
__ver_minor__ = 14
__ver_patch__ = 0
__ver_sub__ = ""
__ver_tuple__ = (__ver_major__, __ver_minor__, __ver_patch__, __ver_sub__)
__version__ = "%d.%d.%d%s" % __ver_tuple__


from browserid.errors import (Error,  # NOQA
                              ConnectionError,  # NOQA
                              TrustError,  # NOAQ
                              ExpiredSignatureError,  # NOQA
                              InvalidSignatureError,  # NOQA
                              AudienceMismatchError)  # NOQA

from browserid.verifiers.remote import RemoteVerifier  # NOQA
from browserid.verifiers.local import LocalVerifier  # NOQA


_DEFAULT_VERIFIER = None


def verify(assertion, audience=None):
    """Verify the given BrowserID assertion.

    This function uses the "best" verification method available in order to
    verify the given BrowserID assertion and return a dict of user data.  The
    best method currently involves POSTing to the hosted verifier service on
    persona.org; eventually it will do local verification.
    """
    global _DEFAULT_VERIFIER
    if _DEFAULT_VERIFIER is None:
        _DEFAULT_VERIFIER = LocalVerifier()
    return _DEFAULT_VERIFIER.verify(assertion, audience)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import time
import hashlib
from contextlib import contextmanager

from browserid.utils import (encode_bytes, bundle_certs_and_assertion,
                             to_int, to_hex)

from browserid import supportdoc
from browserid import jwt

# if unittest2 isn't available, assume that we are python 2.7
try:
    import unittest2 as unittest
except ImportError:
    import unittest  # NOQA


# These are values used to generate dummy DSA keys.
# I took them directly from the javacript jwcrypto source code, which claims:
#    """
#    the following are based on the first FIPS186-3 test vectors for 1024/160
#    SHA-256 under the category A.2.3 Verifiable Canonical Generation of the
#    Generator g
#    """
DUMMY_Q = to_int("e21e04f911d1ed7991008ecaab3bf775984309c3", 16)


DUMMY_P = to_int("""
  ff600483db6abfc5b45eab78594b3533d550d9f1bf2a992a7a8daa6dc34f8045ad4e6e0c429
  d334eeeaaefd7e23d4810be00e4cc1492cba325ba81ff2d5a5b305a8d17eb3bf4a06a349d39
  2e00d329744a5179380344e82a18c47933438f891e22aeef812d69c8f75e326cb70ea000c3f
  776dfdbd604638c2ef717fc26d02e17
""", 16)

DUMMY_G = to_int("""
  c52a4a0ff3b7e61fdf1867ce84138369a6154f4afa92966e3c827e25cfa6cf508b90e5de419
  e1337e07a2e9e2a3cd5dea704d175f8ebf6af397d69e110b96afb17c7a03259329e4829b0d0
  3bbc7896b15b4ade53e130858cc34d96269aa89041f409136c7242a38895c9d5bccad4f389a
  f1d7a4bd1398bd072dffa896233397a
""", 16)


def fetch_support_document(hostname, verify=None):
    """Fetch the BrowserID support document for the given hostname.

    Actually, this implementation generates a key locally based on
    a hash of the hostname.  This lets us exercise all the crypto code
    while using predictable local values.
    """

    if hostname == "redirect.org":
        return {"authority": "delegated.org"}

    if hostname == "redirect-twice.org":
        return {"authority": "redirect.org"}

    if hostname == "infinite.org":
        return {"authority": "infinite.org"}

    return {"public-key": get_keypair(hostname)[0]}


def get_keypair(hostname):
    """Generate a dummy keypair for the given hostname.

    This method generates a dummy DSA keypair for the given hostname.
    It returns a tuple (pub, priv) where "pub" is a dict of values for
    the public key, and "priv" is a DSA128Key object containing the
    private key.  Multiple calls to this method for the same hostname
    are guaranteed to produce the same key.

    To make this work we take advantage of the fact that DSA key generation
    is just "generate x by some random method, where 0 < x < q".  Replace
    "some random method" with "sha1 hash of hostname" and we're all set.
    """
    # Use pre-agreed parameters for p, q and g.
    q = DUMMY_Q
    p = DUMMY_P
    g = DUMMY_G
    # Generate private key x by "some random method".
    x = to_int(hashlib.sha1(hostname.encode("utf8")).hexdigest(), 16)
    assert x != 0, "SHA1(hostname) is zero - what are the odds?!"
    # Calculate public key y as usual.
    y = pow(g, x, p)
    data = {
      "algorithm": "DS",
      "p": to_hex(p),
      "q": to_hex(q),
      "g": to_hex(g),
      "y": to_hex(y),
      "x": to_hex(x),
    }
    privkey = jwt.DS128Key(data)
    del data["x"]
    return data, privkey


def make_assertion(email, audience, issuer=None, exp=None,
                    assertion_sig=None, certificate_sig=None,
                    new_style=True, email_keypair=None, issuer_keypair=None,
                    idp_claims=None, user_claims=None):
    """Generate a new dummy assertion for the given email address.

    This method lets you generate BrowserID assertions using dummy private
    keys. Called with just an email and audience it will generate an assertion
    from login.persona.org.

    By specifying the "exp", "assertion_sig" or "certificate_sig" arguments
    it is possible generate invalid assertions for testing purposes.
    """
    if issuer is None:
        issuer = "login.persona.org"
    if exp is None:
        exp = int((time.time() + 60) * 1000)
    # Get private key for the email address itself.
    if email_keypair is None:
        email_keypair = get_keypair(email)
    email_pub, email_priv = email_keypair
    # Get private key for the hostname so we can sign it.
    if issuer_keypair is None:
        issuer_keypair = get_keypair(issuer)
    iss_pub, iss_priv = issuer_keypair

    # Generate the assertion, signed with email's public key.
    assertion = {
        "exp": exp,
        "aud": audience,
    }
    if user_claims:
        assertion.update(user_claims)
    assertion = jwt.generate(assertion, email_priv)
    if assertion_sig is not None:
        assertion = ".".join(assertion.split(".")[:-1] +
                                [encode_bytes(assertion_sig)])
    # Generate the certificate signing the email's public key
    # with the issuer's public key.
    certificate = {
        "iss": issuer,
        "exp": exp,
        "principal": {"email": email},
        "public-key": email_pub,
    }
    if idp_claims:
        certificate.update(idp_claims)
    certificate = jwt.generate(certificate, iss_priv)
    if certificate_sig is not None:
        certificate = ".".join(certificate.split(".")[:-1] +
                                [encode_bytes(certificate_sig)])
    # Combine them into a BrowserID bundled assertion.
    return bundle_certs_and_assertion([certificate], assertion, new_style)


@contextmanager
def patched_supportdoc_fetching(replacement=None, exc=None):
    """Patch the key fetching mechanism with the given callable.

    This is to allow easier testing.
    """
    def raise_exception(*args, **kwargs):
        raise exc

    if exc is not None:
        replacement = raise_exception
    if replacement is None:
        replacement = fetch_support_document
    old_callable = supportdoc.fetch_support_document
    supportdoc.fetch_support_document = replacement
    yield
    supportdoc.fetch_support_document = old_callable


def callwith(context):
    """Decorator to call a function with a context manager."""
    def decorator(func):
        def wrapper(*args, **kwds):
            with context:
                return func(*args, **kwds)
        return wrapper
    return decorator



#
#  This monkey-patches M2Crypto's RSA and DSA support to allow us
#  to create keys directly from a given set of parameters.
#  It's based on the following patch from the M2Crypt bugtracker:
#
#      https://bugzilla.osafoundation.org/show_bug.cgi?id=12981
#
#  We use ctypes to avoid recompiling the M2Crypto binaries.

import ctypes
from M2Crypto import RSA, DSA, m2, __m2crypto

_m2lib = ctypes.CDLL(__m2crypto.__file__)


_m2lib.BN_free.argtypes = (ctypes.c_void_p,)
_m2lib.BN_mpi2bn.argtypes = (ctypes.c_char_p, ctypes.c_int, ctypes.c_void_p,)
_m2lib.BN_mpi2bn.restype = ctypes.c_void_p


class _RSA(ctypes.Structure):
    """OpenSSL struct representing a RSA key (struct rsa_st)."""
    _fields_ = [("pad", ctypes.c_int),
                ("version", ctypes.c_long),
                ("RSA_METHOD", ctypes.c_void_p),
                ("ENGINE", ctypes.c_void_p),
                ("n", ctypes.c_void_p),
                ("e", ctypes.c_void_p),
                ("d", ctypes.c_void_p)]
                # There are many more fields, but we don't need them.


class _DSA(ctypes.Structure):
    """OpenSSL struct representing a DSA key (struct dsa_st)."""
    _fields_ = [("pad", ctypes.c_int),
                ("version", ctypes.c_long),
                ("write_params", ctypes.c_int),
                ("p", ctypes.c_void_p),
                ("q", ctypes.c_void_p),
                ("g", ctypes.c_void_p),
                ("pub_key", ctypes.c_void_p),
                ("priv_key", ctypes.c_void_p)]
                # There are many more fields, but we don't need them.


def maybe_provide(obj):
    """Decorator to provide default implemenation of a function."""
    def decorator(func):
        if not hasattr(obj, func.__name__):
            setattr(obj, func.__name__, func)
        return func
    return decorator


@maybe_provide(m2)
def rsa_set_d(rsa, value):
    """Set the private-key component "d" of a RSA object."""
    bn = _m2lib.BN_mpi2bn(value, len(value), None)
    if not bn:
        raise RSA.RSAError("invalid private key data")
    rsa_p = ctypes.cast(ctypes.c_void_p(int(rsa)), ctypes.POINTER(_RSA))
    if rsa_p.contents.d:
        _m2lib.BN_free(rsa_p.contents.d)
    rsa_p.contents.d = bn


@maybe_provide(m2)
def dsa_set_pub(dsa, value):
    """Set the public-key component of a DSA object."""
    bn = _m2lib.BN_mpi2bn(value, len(value), None)
    if not bn:
        raise DSA.DSAError("invalid public key data")
    dsa_p = ctypes.cast(ctypes.c_void_p(int(dsa)), ctypes.POINTER(_DSA))
    if dsa_p.contents.pub_key:
        _m2lib.BN_free(dsa_p.contents.pub_key)
    dsa_p.contents.pub_key = bn


@maybe_provide(m2)
def dsa_set_priv(dsa, value):
    """Set the private-key component of a DSA object."""
    bn = _m2lib.BN_mpi2bn(value, len(value), None)
    if not bn:
        raise DSA.DSAError("invalid private key data")
    dsa_p = ctypes.cast(ctypes.c_void_p(int(dsa)), ctypes.POINTER(_DSA))
    if dsa_p.contents.priv_key:
        _m2lib.BN_free(dsa_p.contents.priv_key)
    dsa_p.contents.priv_key = bn


@maybe_provide(RSA)
def new_key(parameters):
    """Create a RSA object from the given parameters."""
    e, n, d = parameters
    rsa = m2.rsa_new()
    m2.rsa_set_e(rsa, e)
    m2.rsa_set_n(rsa, n)
    m2.rsa_set_d(rsa, d)
    return RSA.RSA(rsa, 1)


# Calling sign() on a public key will segfault M2Crypto.
# Stub it out so that it raises an error instead.
if "sign" not in RSA.RSA_pub.__dict__:
    RSA.RSA_pub.sign = RSA.RSA_pub.private_encrypt


@maybe_provide(DSA)
def load_pub_key_params(p, q, g, pub):
    """Create a DSA_pub object from parameters and key."""
    dsa = m2.dsa_new()
    m2.dsa_set_p(dsa, p)
    m2.dsa_set_q(dsa, q)
    m2.dsa_set_g(dsa, g)
    m2.dsa_set_pub(dsa, pub)
    return DSA.DSA_pub(dsa, 1)


@maybe_provide(DSA)
def load_key_params(p, q, g, pub, priv):
    """Create a DSA object from parameters and key."""
    dsa = m2.dsa_new()
    m2.dsa_set_p(dsa, p)
    m2.dsa_set_q(dsa, q)
    m2.dsa_set_g(dsa, g)
    m2.dsa_set_pub(dsa, pub)
    m2.dsa_set_priv(dsa, priv)
    return DSA.DSA(dsa, 1)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import sys
import re
import fnmatch

from browserid.errors import AudienceMismatchError
from browserid.utils import (unbundle_certs_and_assertion,
                             decode_json_bytes)


if sys.version_info > (3,):
    basestring = (str,)


class Verifier(object):
    """Abstract base class for verifying BrowserID assertions."""

    def __init__(self, audiences=None):
        self.audiences = audiences
        self._audience_re = self._compile_audience_patterns(audiences)

    def verify(self, assertion, audience=None):
        """Verify the given BrowserID assertion.

        This method checks the validity of the given BrowserID assertion and,
        if valid, returns the dict of asserted data.  If it is not valid then
        an error will be raised.

        If the 'audience' argument is given, the assertion must be for that
        specific audience.  Otherwise, it must be for an audience matching
        one of the patterns provided in the constructor.
        """
        raise NotImplementedError

    def check_audience(self, assertion, expected_audience=None):
        """Check that the assertion matches the expected audience.

        This method verifies that the audience for the given assertion is
        as expected - either matching the audience parameter if given, or
        or matching one of the audience patterns from the constructor if not.

        If the audience matches then it is returned as a string; if not then
        an AudienceMismatchError is raised.
        """
        try:
            _, token = unbundle_certs_and_assertion(assertion)
            audience = decode_json_bytes(token.split(".")[1])["aud"]
        except (KeyError, IndexError):
            raise ValueError("Malformed JWT")
        if expected_audience is None:
            audience_re = self._audience_re
        else:
            audience_re = self._compile_audience_patterns(expected_audience)
        if audience_re is None:
            raise AudienceMismatchError
        if not audience_re.match(audience):
            raise AudienceMismatchError
        return audience

    def _compile_audience_patterns(self, audiences):
        """Compile a list of audience patterns into a regular expression."""
        if not audiences:
            return None
        if isinstance(audiences, basestring):
            audiences = (audiences,)
        regexps = []
        for pattern in audiences:
            regexp = fnmatch.translate(pattern)
            if "://" not in regexp:
                regexp = "([a-z]+://)?" + regexp
            regexps.append("(" + regexp + ")")
        return re.compile("|".join(regexps))



import json
import socket

from unittest.mock import Mock, patch
from requests.exceptions import RequestException

from browserid.supportdoc import fetch_support_document, SupportDocumentManager
from browserid.errors import ConnectionError, InvalidIssuerError
from browserid.tests.support import unittest
from browserid.tests.support import (fetch_support_document as
        patched_support_document)


# Retrieved from browserid.org on April 3rd 2012
BROWSERID_PK = (
    '{"public-key":{"algorithm":"RS","n":"175097498616944944948724600'
    '5376783505040424585022287992333285107040747762043794156382037076'
    '4990477086508806099113661589535215545809212436371869690217935480'
    '2753014948839838403516326408223252892569346143048785710656475684'
    '9535475273645907806528415369926171343773128172627893352378261396'
    '0153494025694829910802495907763077221584500090734186210456302804'
    '6884323084778492418923884368673543934239778647619964884232166051'
    '7909653959911229288600229842193433562918970249484466937121698566'
    '1583323059605724956419024024496484812121544425787678170853739436'
    '5238417167558546493512404073066199364247440288962324288605736789'
    '20055912798079","e":"65537"}}')
BROWSERID_PK_PY = json.loads(BROWSERID_PK)


class TestFetchPublicKey(unittest.TestCase):

    @patch('browserid.netutils.requests')
    def _fetch(self, hostname, requests, well_known_url=None,
               side_effect=None, response_text='', status_code=200):
        response = Mock()
        response.text = response_text
        response.status_code = status_code
        requests.request.side_effect = side_effect
        requests.request.return_value = response

        kwargs = {}
        if well_known_url is not None:
            kwargs['well_known_url'] = well_known_url

        supportdoc = fetch_support_document(hostname, **kwargs)

        try:
            key = supportdoc['public-key']
        except KeyError:
            raise InvalidIssuerError('Host %r has malformed public key '
                                     'document' % hostname)

        return key

    def test_connection_error(self):
        """If there is an error connecting, raise a ConnectionError."""
        with self.assertRaises(ConnectionError):
            self._fetch('test.com', side_effect=RequestException)
        with self.assertRaises(ConnectionError):
            self._fetch('test.com', side_effect=socket.error)

    @patch('browserid.supportdoc.fetch_support_document')
    def test_missing_support_document(self, fetch):
        with self.assertRaises(InvalidIssuerError):
            self._fetch('test.com', status_code=404)

    def test_malformed_support_document(self):
        response_text = 'I AINT NO JSON, FOOL!'
        with self.assertRaises(InvalidIssuerError):
            self._fetch('test.com', response_text=response_text)

    def test_malformed_pub_key_document(self):
        # We need the first request to raise a 404, so we replace
        # post with a custom function here.
        def post(url, data, verify=True):
            response = Mock()
            if not post.called:
                response.status_code = 404
                post.called = True
            response.text = 'I AINT NO JSON, FOOL!'
            return response
        post.called = False

        with patch('browserid.netutils.requests') as requests:
            requests.request = post
            with self.assertRaises(InvalidIssuerError):
                fetch_support_document('test.com')

    def test_support_document_with_no_public_key(self):
        with self.assertRaises(InvalidIssuerError):
            self._fetch('test.com', response_text='{}')

    def test_successful_fetch(self):
        key = self._fetch('test.com', response_text=BROWSERID_PK)
        self.assertEquals(key, BROWSERID_PK_PY['public-key'])


class TestTrustedIssuers(unittest.TestCase):
    def setUp(self):
        self.supportdocmgr = SupportDocumentManager()

    def _is_trusted_issuer(self, *args, **kwds):
        return self.supportdocmgr.is_trusted_issuer(*args, **kwds)

    def test_trusted_secondaries(self):
        PERSONA_HOST = 'login.persona.org'
        self.assertTrue(self._is_trusted_issuer('test.com', PERSONA_HOST))
        self.assertFalse(self._is_trusted_issuer('test.com', PERSONA_HOST,
            trusted_secondaries=[], max_delegations=0))

    def test_hostname_issuer(self):
        self.assertTrue(self._is_trusted_issuer('test.com', 'test.com'))
        self.assertFalse(self._is_trusted_issuer('abc.com', 'test.com',
            max_delegations=0))

    @patch('browserid.supportdoc.fetch_support_document',
           patched_support_document)
    def test_delegated_primary(self):
        self.assertTrue(self._is_trusted_issuer('redirect.org',
            'delegated.org'))

    def test_disabled_delegated_primary(self):
        self.assertFalse(self._is_trusted_issuer('redirect.org',
            'delegated.org', max_delegations=0))

    @patch('browserid.supportdoc.fetch_support_document',
           patched_support_document)
    def test_infinite_delegated_primary_recursion(self):
        self.assertFalse(self._is_trusted_issuer('infinite.org', None))
        self.assertFalse(self._is_trusted_issuer('infinite.org',
                         'delegated.org'))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

from browserid.tests.support import unittest
from browserid.utils import to_int

try:
    from browserid.crypto.m2 import int2mpint
    import browserid.crypto._m2_monkeypatch as _m2
    HAVE_M2CRYPTO = True
except ImportError:
    HAVE_M2CRYPTO = False


# Dummy RSA key for testing purposes.

DUMMY_RSA_E = to_int("65537")
DUMMY_RSA_N = to_int("""110897663942528265066856163966583557538666146275146
                     569193074111045116764854772535689458732714049671807506
                     396649306730328647317126800964431366624486416551078177
                     528195103050868728550429561392842977259407335332582178
                     624191611001106449477645116630750398871838788574825885
                     770446686329706009000279629721965986677219""")
DUMMY_RSA_D = to_int("""295278123166626215026113502482091502365034141401240
                     159363282304307076544046230487782634982660202141239450
                     481640966544735782181647417005558287318200095948234745
                     214183393770321992676297531378428617531522265932631860
                     693144704788708252936752025413728425562033678747736289
                     64114133156747686886305629893015763517873""")


class TestM2MonkeyPatch(unittest.TestCase):

    def setUp(self):
        super(TestM2MonkeyPatch, self).setUp()
        if not HAVE_M2CRYPTO:
            raise unittest.SkipTest()

    def test_setting_invalid_data_on_dsa_key(self):
        k = _m2.DSA.gen_params(512)
        k.gen_key()
        _m2.dsa_set_pub(k.dsa, k.pub)
        self.assertRaises(_m2.DSA.DSAError, _m2.dsa_set_pub, k.dsa, "\x00")
        _m2.dsa_set_priv(k.dsa, k.priv)
        self.assertRaises(_m2.DSA.DSAError, _m2.dsa_set_priv, k.dsa, "\x00")

    def test_setting_invalid_data_on_rsa_key(self):
        args = map(int2mpint, (DUMMY_RSA_E, DUMMY_RSA_N, DUMMY_RSA_D))
        k = _m2.RSA.new_key(args)
        self.assertTrue(k.verify("hello", k.sign("hello")))
        _m2.rsa_set_d(k.rsa, int2mpint(DUMMY_RSA_D))
        self.assertRaises(_m2.RSA.RSAError, _m2.rsa_set_d, k.rsa, "\x00")
        self.assertTrue(k.verify("hello", k.sign("hello")))

    def test_dsa_signing_works_with_loaded_keys(self):
        d_orig = _m2.DSA.gen_params(512)
        d_orig.gen_key()
        d_pub = _m2.DSA.load_pub_key_params(d_orig.p, d_orig.q, d_orig.g,
                                        d_orig.pub)
        d_priv = _m2.DSA.load_key_params(d_orig.p, d_orig.q, d_orig.g,
                                     d_orig.pub, d_orig.priv)
        # Check that the attributes are copied across effectively.
        for nm in ("p", "q", "g", "pub"):
            self.assertEquals(getattr(d_orig, nm), getattr(d_pub, nm))
            self.assertEquals(getattr(d_orig, nm), getattr(d_priv, nm))
        self.assertEquals(d_orig.priv, d_priv.priv)
        # Check that they can all validate signatures from original key.
        r, s = d_orig.sign("helloworld")
        self.assertTrue(d_orig.verify("helloworld", r, s))
        self.assertTrue(d_pub.verify("helloworld", r, s))
        self.assertTrue(d_priv.verify("helloworld", r, s))
        # Check that they can all validate signatures from loaded priv key.
        r, s = d_priv.sign("helloworld")
        self.assertTrue(d_orig.verify("helloworld", r, s))
        self.assertTrue(d_pub.verify("helloworld", r, s))
        self.assertTrue(d_priv.verify("helloworld", r, s))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""

Utilities for dealing with Signed JSON Web Tokens.

"""

import hashlib

from browserid.crypto import Key, DSKey, RSKey  # NOQA

from browserid.utils import decode_bytes, encode_bytes
from browserid.utils import decode_json_bytes, encode_json_bytes


def parse(jwt):
    """Parse a JWT from a string."""
    algorithm, payload, signature = jwt.split(".")
    signed_data = algorithm + "." + payload
    try:
        algorithm = decode_json_bytes(algorithm)["alg"]
    except KeyError:
        raise ValueError("badly formed JWT")
    payload = decode_json_bytes(payload)
    signature = decode_bytes(signature)
    return JWT(algorithm, payload, signature, signed_data)


def generate(payload, key):
    """Generate and sign a JWT for a dict payload."""
    alg = key.__class__.__name__[:-3]
    algorithm = encode_json_bytes({"alg": alg})
    payload = encode_json_bytes(payload)
    signed_data = ".".join((algorithm, payload)).encode("ascii")
    signature = encode_bytes(key.sign(signed_data))
    return ".".join((algorithm, payload, signature))


class JWT(object):
    """Class for parsing signed JSON Web Tokens.

    To parse a JWT from a bytestring, use the module-level parse() function.
    This class is really only for internal purposes.
    """

    def __init__(self, algorithm, payload, signature, signed_data):
        self.algorithm = algorithm
        self.payload = payload
        self.signature = signature
        self.signed_data = signed_data

    def check_signature(self, key_data):
        """Check that the JWT was signed with the given key."""
        if not self.algorithm.startswith(key_data["algorithm"]):
            return False
        key = load_key(self.algorithm, key_data)
        return key.verify(self.signed_data.encode("ascii"), self.signature)


def load_key(algorithm, key_data):
    """Load a Key object from the given data."""
    if not algorithm.isalnum():
        msg = "unknown signing algorithm: %s" % (algorithm,)
        raise ValueError(msg)
    try:
        key_class = globals()[algorithm + "Key"]
    except KeyError:
        msg = "unknown signing algorithm: %s" % (algorithm,)
        raise ValueError(msg)
    return key_class(key_data)


class RS64Key(RSKey):
    DIGESTSIZE = 256
    HASHNAME = "sha256"
    HASHMOD = hashlib.sha256


class RS128Key(RSKey):
    DIGESTSIZE = 320
    HASHNAME = "sha256"
    HASHMOD = hashlib.sha256


class RS256Key(RSKey):
    DIGESTSIZE = 512
    HASHNAME = "sha256"
    HASHMOD = hashlib.sha256


class DS128Key(DSKey):
    BITLENGTH = 160
    HASHMOD = hashlib.sha1


class DS256Key(DSKey):
    BITLENGTH = 256
    HASHMOD = hashlib.sha256



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import sys
import json
import collections
import threading
import time

if sys.version_info > (3,):
    from urllib.parse import urljoin
else:
    from urlparse import urljoin  # NOQA

from browserid import netutils
from browserid.errors import (ConnectionError,
                              InvalidIssuerError)

WELL_KNOWN_URL = "/.well-known/browserid"

DEFAULT_MAX_DELEGATIONS = 6

DEFAULT_TRUSTED_SECONDARIES = ("browserid.org", "diresworb.org",
                               "dev.diresworb.org" "login.anosrep.org",
                               "login.persona.org")


class SupportDocumentManager(object):
    """Manager for mapping hostnames to their BrowserID support documents.

    This object handles the association of a hostname with its BrowserID
    support document, if any.  It automatically fetches support documents
    as required and stores them in a cache for future reference.
    """

    def __init__(self, cache=None, verify=None, **kwargs):
        self.verify = verify
        if cache is None:
            cache = FIFOCache(**kwargs)
        self.cache = cache

    def get_support_document(self, hostname):
        """Get the BrowserID support document for the given hostname."""
        try:
            # Use a cached key if available.
            (error, supportdoc) = self.cache[hostname]
        except KeyError:
            # Fetch the key afresh from the specified server.
            # Cache any failures so we're not flooding bad hosts.
            error = supportdoc = None
            try:
                supportdoc = self.fetch_support_document(hostname)
            except Exception as e:  # NOQA
                error = e
            self.cache[hostname] = (error, supportdoc)
        if error is not None:
            raise error
        return supportdoc

    def get_key(self, hostname):
        """Get the public key for verifying assertions from the given host."""
        supportdoc = self.get_support_document(hostname)
        try:
            key = supportdoc['public-key']
        except KeyError:
            raise InvalidIssuerError(
                    "Host %r doesn't provide a public key" % hostname)

        return key

    def fetch_support_document(self, hostname):
        """Fetch the BrowserID support document for the given hostname."""
        return fetch_support_document(hostname, verify=self.verify)

    def is_trusted_issuer(self, hostname, issuer, trusted_secondaries=None,
            max_delegations=DEFAULT_MAX_DELEGATIONS):
        """Check whether an issuer is trusted to assert for a given hostname.

        This method checks whether the given issuer is trusted to assert
        identitiers for the given hostname.  There are three cases in which
        this can be true:

          * The hostname and issuer and the same
          * The issuer is in the list of trusted secondaries
          * The hostname delegates authority to the issuer

        You can disable the check for delegated authority by setting the
        max_delegations argument to 0.
        """
        if hostname == issuer:
            return True

        if trusted_secondaries is None:
            trusted_secondaries = DEFAULT_TRUSTED_SECONDARIES

        if issuer in trusted_secondaries:
            return True

        num_delegations = 0
        while num_delegations < max_delegations:
            try:
                supportdoc = self.get_support_document(hostname)
            except (InvalidIssuerError, ConnectionError):
                break
            authority = supportdoc.get("authority")
            if authority is None:
                break
            if authority == issuer:
                return True
            hostname = authority
            num_delegations += 1

        return False


class FIFOCache(object):
    """A simple in-memory FIFO cache for BrowserID public keys.

    This is a *very* simple in-memory FIFO cache, used as the default object
    for caching BrowserID public keys in the LocalVerifier.  Items are kept
    for 'cache_timeout' seconds before being evicted from the cache.  If the
    'max_size' argument is not None and the cache grows above this size,
    items will be evicted early in order of insertion into the cache.

    (An LFU cache would be better but that's a whole lot more work...)
    """

    def __init__(self, cache_timeout=60 * 60, max_size=1000):
        self.cache_timeout = cache_timeout
        self.max_size = max_size
        self.items_map = {}
        self.items_queue = collections.deque()
        self._lock = threading.Lock()

    def __getitem__(self, key):
        """Lookup the given key in the cache.

        This method retrieves the value cached under the given key, evicting
        it from the cache if expired.
        """
        (timestamp, value) = self.items_map[key]
        if self.cache_timeout:
            expiry_time = timestamp + self.cache_timeout
            if expiry_time < time.time():
                # Lock the cache while evicting, and double-check that
                # it hasn't been updated by another thread in the meantime.
                # This is a little more work during eviction, but it means we
                # can avoid locking in the common case of non-expired items.
                self._lock.acquire()
                try:
                    if self.items_map[key][0] == timestamp:
                        # Just delete it from items_map.  Trying to find
                        # and remove it from items_queue would be expensive,
                        # so we count on a subsequent write to clean it up.
                        del self.items_map[key]
                except KeyError:
                    pass
                finally:
                    self._lock.release()
                    raise KeyError
        return value

    def __setitem__(self, key, value):
        """Cache the given value under the given key.

        This method caches the given value under the given key, checking that
        there's enough room in the cache and evicting items if necessary.
        """
        now = time.time()
        with self._lock:
            # First we need to make sure there's enough room.
            # This is a great opportunity to evict any expired items,
            # helping to keep memory small for sparse caches.
            if self.cache_timeout:
                expiry_time = now - self.cache_timeout
                while self.items_queue:
                    (e_key, e_item) = self.items_queue[0]
                    if e_item[0] >= expiry_time:
                        break
                    self.items_queue.popleft()
                    if self.items_map.get(e_key) == e_item:
                        del self.items_map[e_key]
            # If the max size has been exceeded, evict things in time order.
            if self.max_size:
                while len(self.items_map) >= self.max_size:
                    (e_key, e_item) = self.items_queue.popleft()
                    if self.items_map.get(e_key) == e_item:
                        del self.items_map[e_key]
            # Now we can store the incoming item.
            item = (now, value)
            self.items_queue.append((key, item))
            self.items_map[key] = item

    def __delitem__(self, key):
        """Remove the given key from the cache."""
        # This is a lazy delete.  Removing it from items_map means it
        # wont be found by __get__, and the entry in items_queue will
        # get cleaned up when its expiry time rolls around.
        del self.items_map[key]

    def __len__(self):
        """Get the currently number of items in the cache."""
        return len(self.items_map)


def fetch_support_document(hostname, well_known_url=None, verify=None):
    """Fetch the BrowserID well-known file for the given hostname.

    This function fetches and parses the well-known BrowserID meta-data file.

    :param verify: verify the certificate when requesting ssl resources
    """
    if well_known_url is None:
        well_known_url = WELL_KNOWN_URL

    hostname = 'https://%s' % hostname

    # Try to find the support document.  If it can't be found then we
    # raise an InvalidIssuerError.  Any other connection-related
    # errors are passed back up to the caller.
    response = netutils.get(urljoin(hostname, well_known_url), verify=verify)
    if response.status_code == 200:
        try:
            data = json.loads(response.text)
        except ValueError:
            raise InvalidIssuerError('Host %r has malformed BrowserID '
                                     'support document' % hostname)
    else:
        # The well-known file was not found, try falling back to
        # just "/pk".
        response = netutils.get(urljoin(hostname, '/pk'), verify=verify)
        if response.status_code == 200:
            try:
                key = json.loads(response.text)
            except ValueError:
                raise InvalidIssuerError('Host %r has malformed BrowserID '
                                         'metadata document' % hostname)

            data = {"public-key": key}
        else:
            raise InvalidIssuerError('Host %r does not declare support for '
                                     'BrowserID' % hostname)

    return data



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""

Utility functions for PyBrowserID.

"""

import sys
import json
import base64


if sys.version_info > (3,):
    long = int
    unicode = str


def decode_bytes(value):
    """Decode BrowserID's base64 encoding format.

    BrowserID likes to strip padding characters off of base64-encoded strings,
    meaning we can't use the stdlib routines to decode them directly.  This
    is a simple wrapper that adds the padding back in.

    If the value is not correctly encoded, ValueError will be raised.
    """
    if isinstance(value, unicode):
        value = value.encode("ascii")
    pad = len(value) % 4
    if pad == 2:
        value += b"=="
    elif pad == 3:
        value += b"="
    elif pad != 0:
        raise ValueError("incorrect b64 encoding")
    try:
        return base64.urlsafe_b64decode(value)
    except TypeError as e:
        raise ValueError(str(e))


def encode_bytes(value):
    """Encode BrowserID's base64 encoding format.

    BrowserID likes to strip padding characters off of base64-encoded strings,
    meaning we can't use the stdlib routines to encode them directly.  This
    is a simple wrapper that strips the padding.
    """
    if isinstance(value, unicode):
        value = value.encode("ascii")
    return base64.urlsafe_b64encode(value).rstrip(b"=").decode("ascii")


def decode_json_bytes(value):
    """Decode a JSON object from some encoded bytes.

    This function decodes a JSON object from bytes encoded in the BrowserID
    base64 format.  If the bytes and invalid or do not encode a proper object,
    ValueError is raised.
    """
    obj = json.loads(decode_bytes(value).decode("utf8"))
    if not isinstance(obj, dict):
        raise ValueError("JSON did not contain an object")
    return obj


def encode_json_bytes(obj):
    """Encode an object as JSON bytes in the BrowserID base64 format."""
    if not isinstance(obj, dict):
        raise ValueError("Value is not a JSON object")
    return encode_bytes(json.dumps(obj).encode("utf8"))


def bundle_certs_and_assertion(certificates, assertion, new_style=True):
    """Bundle certificates and assertion into a single string.

    This function produces a BrowserID "bundled assertion" that combines the
    certificate chain and final assertion into a single string.  By default
    it uses the "new-style" tilde-separated format; pass new_style=False to
    use the older b64-encoded-JSON format.
    """
    if new_style:
        return "~".join(certificates) + "~" + assertion
    else:
        return encode_json_bytes({
          "certificates": certificates,
          "assertion": assertion,
        })


def unbundle_certs_and_assertion(bundle):
    """Unbundle certificates and assertion from a single string.

    This function parses a BrowserID "bundled assertion" into the contained
    chain of certificates and final assertion.  The returned value is a tuple
    (certificates, assertion).
    """
    if "~" in bundle:
        certificates, assertion = bundle.rsplit("~", 1)
        certificates = certificates.split("~")
    else:
        data = decode_json_bytes(bundle)
        certificates = data["certificates"]
        assertion = data["assertion"]
    return certificates, assertion


def get_assertion_info(assertion):
    """Parse interesting information out of a BrowserID assertion.

    This function decodes and parses the given BrowserID assertion, returning
    a dict with the following items:

       * principal:  the asserted identity, eg: {"email": "test@example.com"}
       * audience:   the audience to whom it is asserted

    This does *not* verify the assertion at all, it is merely a way to see
    the information that is being asserted.  If the assertion is malformed
    then ValueError will be raised.
    """
    info = {}
    try:
        certificates, assertion = unbundle_certs_and_assertion(assertion)
        # Get the asserted principal out of the certificate chain.
        payload = decode_json_bytes(certificates[-1].split(".")[1])
        info["principal"] = payload["principal"]
        # Get the audience out of the assertion token.
        payload = decode_json_bytes(assertion.split(".")[1])
        info["audience"] = payload["aud"]
    except (TypeError, KeyError) as e:
        raise ValueError(e)
    return info


def to_int(value, base=10):
    """Convert the given value to a python integer.

    The given value can be an existing int or long object, or a string in
    the given base.  The result will always be a long on python2 and an
    int on python3 (which has not concept of a separate "long" type).
    """
    if not isinstance(value, str):
        return long(value)
    return long(value.replace(" ", "").replace("\n", "").strip(), base)


def to_hex(value):
    """Convert the given value to a long encoded into a hex string."""
    return hex(to_int(value))[2:].rstrip("L")


def u(value):
    """Helper function for constructing unicode string literals.

    Use it in lieu of the u"" prefix, like this:

        data = u("unicode string")

    """
    if sys.version_info < (3,):
        value = value.decode("unicode-escape")
    return value
